"""Import everything from functools and some custom functions."""
import inspect
import numbers
from functools import *
from typing import Callable, Iterator, Any, TypeVar, Generic, Tuple

try:
    from typing import ParamSpec
except ImportError:
    from typing_extensions import ParamSpec

_C = TypeVar('_C', bound=Callable)
_T = TypeVar('_T', bound=numbers.Complex)
_S = TypeVar('_S')
_P = ParamSpec('_P')


class FunctionChain(Generic[_C]):
    """Chain of multiple functions. The return value(s) of each previous
    function are the first ``n_args`` positional argument(s) of the next
    function call. This class is for convenient reuse of function chains
    and passing them around as callable objects.

    Similar to :func:`functools.reduce`, but instead of reducing a sequence using
    a single function, this reduces a list of functions by applying them
    iteratively to the output of the function before.

    Example:

    >>> from qutil.functools import chain
    >>> import numpy as np
    >>> x = np.array([1, 4, -6, 8], dtype=float)
    >>> f_chain = FunctionChain(np.abs, np.sqrt)
    >>> f_chain(x, out=x)  # Will write all intermediate results into the same array.
    array([1.        , 2.        , 2.44948974, 2.82842712])

    n_args argument:

    >>> def adder(x, axis):
    ...     return x.sum(axis), axis - 1
    >>> def multiplier(x, axis):
    ...     return x.prod(axis), axis - 1
    >>> x = np.arange(12).reshape(3, 4)
    >>> axis = 1
    >>> f_chain = FunctionChain(adder, multiplier, n_args=2)
    >>> result, axis = f_chain(x, axis)
    >>> print(result)
    5016
    """
    __array_interface__ = {
        'shape': (),
        'typestr': '|O',
        'version': 3
    }
    """Describes to NumPy how to convert this object into an array."""

    def __init__(self, *functions: _C, n_args: int = 1, inspect_kwargs: bool = False):
        self.functions = functions
        self.n_args = n_args
        self.inspect_kwargs = inspect_kwargs
        if self.n_args < 1:
            raise ValueError(f'n_args should be a positive integer, not {n_args}.')

    def __getitem__(self, item: Any):
        raise TypeError('The __getitem__ method has been removed. Index into '
                        'FunctionChain.functions instead')

    def __len__(self) -> int:
        return len(self.functions)

    def __iter__(self) -> Iterator[_C]:
        raise TypeError('The __iter__ method has been removed. Iterate over '
                        'FunctionChain.functions instead')

    def __repr__(self) -> str:
        return (
            super().__repr__()
            + ' with functions'
            + ('\n - {}'*len(self)).format(*self.functions)
        )

    def __call__(self, *args, **kwargs):
        """Iteratively apply functions to the return value of the
        previous one, starting with `x` as initial argument.

        Args:
            *args: Positional arguments that get passed to each function besides the previous function's return value.

            **kwargs: Keyword arguments that get passed to each function.

        Returns:
            Return value of the last function
        """
        args = list(args)
        for func in self.functions:
            if self.inspect_kwargs:
                tmp_kwargs, _ = filter_kwargs(func, kwargs)
            else:
                tmp_kwargs = kwargs
            if self.n_args > 1:
                args[:self.n_args] = func(*args, **tmp_kwargs)
            else:
                args[0] = func(args[0], *args[1:], **tmp_kwargs)
        return tuple(args[:self.n_args]) if self.n_args > 1 else args[0]


def chain(*functions: _C, n_args: int = 1, inspect_kwargs: bool = False) -> FunctionChain[_C]:
    """Chain multiple functions.

    The return value of each previous function is the first argument of
    the next function call.

    Example:
        >>> from qutil.functools import chain
        >>> import numpy as np
        >>> f_chain = chain(np.diff, np.sum, print)
        >>> f_chain([1, 3, 6])
        5

    Args:
        *functions : callable
            Functions to be chained.

        n_args : int
            Number of arguments accepted and returned by functions.

        inspect_kwargs : bool
            If true, only pass kwargs actually accepted by the functions.
            Inspects the signature and therefore has some overhead.

    Returns:
        Callable object that chains arguments.
    """
    return FunctionChain(*functions, n_args=n_args, inspect_kwargs=inspect_kwargs)


def scaled(scale: numbers.Complex) -> Callable:
    """Wraps a function; multiplies its result with ``scale``.

    Examples
    --------
    >>> mega = scaled(1e6)(lambda x: x)
    >>> mega(2)
    2000000.0
    >>> @scaled(0.5)
    ... def mysum(a, b):
    ...     return a + b
    >>> mysum(2, 1.5j)
    (1+0.75j)
    """

    def wrapper(fun: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(fun)
        def wrapped(*args: _P.args, **kwargs: _P.kwargs):
            return scale * fun(*args, **kwargs)

        return wrapped

    return wrapper


def filter_kwargs(obj: Callable[_P, _T], kwargs: _P.kwargs) -> Tuple[Any, Any]:
    r"""Inspect `obj`\s signature and filter the set of kwargs into ones
    that `obj` accepts and remaining.

    Examples
    --------
    >>> def foo(x):
    ...     return x*2
    >>> def bar(a, y=4):
    ...     return a + y/2
    >>> def baz(**kw):
    ...     foo_kw, bar_kw = filter_kwargs(foo, kw)
    ...     return bar(foo(**foo_kw), **bar_kw)
    >>> baz(x=2, y=4)
    6.0

    """
    sig = inspect.signature(obj)
    accepted = {}
    remaining = {}
    for key, val in kwargs.items():
        if (
                key in sig.parameters
                and (param := sig.parameters[key]).kind in (
                    param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD
                )
        ):
            accepted[key] = val
        else:
            remaining[key] = val

    return accepted, remaining


del Iterator, Any, TypeVar, ParamSpec
del numbers



"""
Tools for easier interoparability with QCoDeS and qtt
"""
from typing import Mapping, Callable

import numpy as np
import pandas as pd

try:
    from qcodes_loop.data.data_set import DataSet, new_data
    from qcodes_loop.data.data_array import DataArray
except ImportError:
    # legacy qcodes version
    from qcodes.data.data_set import DataSet, new_data
    from qcodes.data.data_array import DataArray

DEFAULT_TO_MV_TRAFO = {
    'V': lambda x: x * 1e3,
    'uV': lambda x: x * 1e-3,
    'µV': lambda x: x * 1e-3
}


def dataframe_to_legacy_dataset(df: pd.DataFrame, unitlookup: Mapping[str, str]) -> DataSet:
    """Create a DataSet from given DataFrame. Set arrays are constructed from
    the index which needs to be a MultiIndex. metadata is copied from df.attrs.
    default_parameter_name is the forst column.

    Parameters
    ----------
    df : pd.DataFrame
        pandas DataFrame with df.index being a pandas.MultiIndex.
    unitlookup : TYPE
        Units of columns and index levels are looked up here.
        All must be present.

    Returns
    -------
    data : DataSet
        default_parameter_name.

    """
    data = new_data()
    data.metadata.update(df.attrs)

    assert isinstance(df.index, pd.MultiIndex)

    das = []
    for lvl, lvldata in enumerate(df.index.levels):
        levname = df.index.names[lvl]
        levdata = df.index.get_level_values(lvl).values.reshape(df.index.levshape)
        levidx = (slice(None, None),) * (lvl + 1) + (0,) * (len(df.index.levshape) - (lvl + 1))
        levdata = levdata[levidx]

        da = DataArray(name=lvldata.name,
                  unit=unitlookup[lvldata.name],
                  array_id=lvldata.name,
                  shape=levdata.shape,
                  is_setpoint=True,
                  preset_data=levdata)
        das.append(da)
        data.add_array(da)

    for colname, colser in df.items():
        col_data = colser.values.reshape(df.index.levshape)

        da = DataArray(name=colname,
          unit=unitlookup[colname],
          array_id=colname,
          shape=col_data.shape,
          is_setpoint=False,
          set_arrays=das,
          preset_data=col_data)
        data.add_array(da)
    data.metadata['default_parameter_name'] = df.columns[0]
    return data


def to_mv_dataset(data: DataSet,
                  to_mv_trafo: Mapping[str, Callable[[np.ndarray], np.ndarray]] = None) -> DataSet:
    """
    Convert all voltage units to mV using the transformations given in
    to_mv_trafo. Some qtt function expect x and y axis to be mV.

    Parameters
    ----------
    data : DataSet
        DESCRIPTION.
    to_mv_trafo : Mapping[str, Callable[[np.ndarray], np.ndarray]], optional
        mapping of units to trafo functions. The default is None which gets
        replaced with DEFAULT_TO_MV_TRAFO.

    Returns
    -------
    DataSet
        Dataset with all voltage like units beeing millivolt.

    """
    if to_mv_trafo is None:
        to_mv_trafo = DEFAULT_TO_MV_TRAFO

    if all(array.unit not in to_mv_trafo for array in data.arrays.values()):
        return data

    new_dataset = new_data()
    new_dataset.metadata.update(data.metadata)

    for array in data.arrays.values():
        if array.is_setpoint:
            if array.set_arrays:
                # a setpoint array must be its own inner loop
                pass

            new_dataset.add_array(DataArray(
                name=array.name,
                array_id=array.name,
                shape=array.shape,
                label=array.label,
                full_name=array.full_name,
                is_setpoint=True,
                unit='mV',
                preset_data=to_mv_trafo.get(array.unit, lambda x: x)(array.ndarray)
                ))

    for array in data.arrays.values():
        if not array.is_setpoint:
            set_arrays = tuple(new_dataset.arrays[set_array.name]
                               for set_array in array.set_arrays)

            new_dataset.add_array(DataArray(
                name=array.name,
                array_id=array.name,
                shape=array.shape,
                label=array.label,
                full_name=array.full_name,
                is_setpoint=False,
                set_arrays=set_arrays,
                unit='mV',
                preset_data=to_mv_trafo.get(array.unit, lambda x: x)(array.ndarray)
                ))

    return new_dataset
